'use strict';

Object.defineProperty(exports, '__esModule', { value: true });

var hash = require('@emotion/hash');

function _interopDefaultLegacy (e) { return e && typeof e === 'object' && 'default' in e ? e : { 'default': e }; }

var hash__default = /*#__PURE__*/_interopDefaultLegacy(hash);

function ampCount(selector) {
    let cnt = 0;
    for (let i = 0; i < selector.length; ++i) {
        if (selector[i] === '&')
            ++cnt;
    }
    return cnt;
}
/**
 * Don't just use ',' to separate css selector. For example:
 * x:(a, b) {} will be split into 'x:(a' and 'b)', which is not expected.
 * Make sure comma doesn't exist inside parentheses.
 */
const seperatorRegex = /\s*,(?![^(]*\))\s*/g;
const extraSpaceRegex = /\s+/g;
/**
 * selector must includes '&'
 * selector is trimmed
 * every part of amp is trimmed
 */
function resolveSelectorWithAmp(amp, selector) {
    const nextAmp = [];
    selector.split(seperatorRegex).forEach(partialSelector => {
        let round = ampCount(partialSelector);
        // eslint-disable-next-line @typescript-eslint/strict-boolean-expressions
        if (!round) {
            amp.forEach(partialAmp => {
                nextAmp.push(
                // eslint-disable-next-line @typescript-eslint/strict-boolean-expressions
                (partialAmp && partialAmp + ' ') + partialSelector);
            });
            return;
        }
        else if (round === 1) {
            amp.forEach(partialAmp => {
                nextAmp.push(partialSelector.replace('&', partialAmp));
            });
            return;
        }
        let partialNextAmp = [
            partialSelector
        ];
        // eslint-disable-next-line @typescript-eslint/strict-boolean-expressions
        while (round--) {
            const nextPartialNextAmp = [];
            partialNextAmp.forEach(selectorItr => {
                amp.forEach(partialAmp => {
                    nextPartialNextAmp.push(selectorItr.replace('&', partialAmp));
                });
            });
            partialNextAmp = nextPartialNextAmp;
        }
        partialNextAmp.forEach(part => nextAmp.push(part));
    });
    return nextAmp;
}
/**
 * selector mustn't includes '&'
 * selector is trimmed
 */
function resolveSelector(amp, selector) {
    const nextAmp = [];
    selector.split(seperatorRegex).forEach(partialSelector => {
        amp.forEach(partialAmp => {
            // eslint-disable-next-line @typescript-eslint/strict-boolean-expressions
            nextAmp.push(((partialAmp && partialAmp + ' ') + partialSelector));
        });
    });
    return nextAmp;
}
function parseSelectorPath(selectorPaths) {
    let amp = [''];
    selectorPaths.forEach(selector => {
        // eslint-disable-next-line
        selector = selector && selector.trim();
        if (
        // eslint-disable-next-line @typescript-eslint/strict-boolean-expressions
        !selector) {
            /**
             * if it's a empty selector, do nothing
             */
            return;
        }
        // eslint-disable-next-line @typescript-eslint/strict-boolean-expressions
        if (selector.includes('&')) {
            amp = resolveSelectorWithAmp(amp, selector);
        }
        else {
            amp = resolveSelector(amp, selector);
        }
    });
    return amp.join(', ').replace(extraSpaceRegex, ' ');
}

const kebabRegex = /[A-Z]/g;
function kebabCase(pattern) {
    return pattern.replace(kebabRegex, match => '-' + match.toLowerCase());
}
/** TODO: refine it to solve nested object */
function upwrapProperty(prop, indent = '  ') {
    if (typeof prop === 'object' && prop !== null) {
        return (' {\n' +
            Object.entries(prop).map(v => {
                return indent + `  ${kebabCase(v[0])}: ${v[1]};`;
            }).join('\n') +
            '\n' + indent + '}');
    }
    return `: ${prop};`;
}
/** unwrap properties */
function upwrapProperties(props, instance, params) {
    if (typeof props === 'function') {
        return props({
            context: instance.context,
            props: params
        });
    }
    return props;
}
function createStyle(selector, props, instance, params) {
    // eslint-disable-next-line @typescript-eslint/strict-boolean-expressions
    if (!props)
        return '';
    // eslint-disable-next-line
    const unwrappedProps = upwrapProperties(props, instance, params);
    // eslint-disable-next-line @typescript-eslint/strict-boolean-expressions
    if (!unwrappedProps)
        return '';
    if (typeof unwrappedProps === 'string') {
        return `${selector} {\n${unwrappedProps}\n}`;
    }
    const propertyNames = Object.keys(unwrappedProps);
    if (propertyNames.length === 0) {
        // eslint-disable-next-line @typescript-eslint/strict-boolean-expressions
        if (instance.config.keepEmptyBlock)
            return selector + ' {\n}';
        return '';
    }
    // eslint-disable-next-line @typescript-eslint/strict-boolean-expressions
    const statements = selector ? [
        selector + ' {'
    ] : [];
    propertyNames.forEach(propertyName => {
        const property = unwrappedProps[propertyName];
        if (propertyName === 'raw') {
            statements.push('\n' + property + '\n');
            return;
        }
        propertyName = kebabCase(propertyName);
        // eslint-disable-next-line @typescript-eslint/strict-boolean-expressions
        if (property !== null && property !== undefined) {
            statements.push(`  ${propertyName}${upwrapProperty(property)}`);
        }
    });
    // eslint-disable-next-line @typescript-eslint/strict-boolean-expressions
    if (selector) {
        statements.push('}');
    }
    return statements.join('\n');
}
function loopCNodeListWithCallback(children, options, callback) {
    /* istanbul ignore if */
    // eslint-disable-next-line @typescript-eslint/strict-boolean-expressions
    if (!children)
        return;
    children.forEach(child => {
        if (Array.isArray(child)) {
            loopCNodeListWithCallback(child, options, callback);
        }
        else if (typeof child === 'function') {
            const grandChildren = child(options);
            if (Array.isArray(grandChildren)) {
                loopCNodeListWithCallback(grandChildren, options, callback);
                // eslint-disable-next-line @typescript-eslint/strict-boolean-expressions
            }
            else if (grandChildren) {
                callback(grandChildren);
            }
            // eslint-disable-next-line @typescript-eslint/strict-boolean-expressions
        }
        else if (child) {
            callback(child);
        }
    });
}
function traverseCNode(node, selectorPaths, styles, instance, params) {
    const $ = node.$;
    // eslint-disable-next-line @typescript-eslint/strict-boolean-expressions
    if (!$ || typeof $ === 'string') {
        // as a string selector
        selectorPaths.push($);
    }
    else if (typeof $ === 'function') {
        // as a lazy selector
        selectorPaths.push($({
            context: instance.context,
            props: params
        }));
    }
    else { // as a option selector
        // eslint-disable-next-line @typescript-eslint/strict-boolean-expressions
        if ($.before)
            $.before(instance.context);
        // eslint-disable-next-line @typescript-eslint/strict-boolean-expressions
        if (!$.$ || typeof $.$ === 'string') {
            selectorPaths.push($.$);
            // eslint-disable-next-line @typescript-eslint/strict-boolean-expressions
        }
        else /* istanbul ignore else */ if ($.$) {
            selectorPaths.push($.$({
                context: instance.context,
                props: params
            }));
        }
    }
    const selector = parseSelectorPath(selectorPaths);
    const style = createStyle(selector, node.props, instance, params);
    // eslint-disable-next-line @typescript-eslint/strict-boolean-expressions
    if (style.length)
        styles.push(style);
    // eslint-disable-next-line @typescript-eslint/strict-boolean-expressions
    if (node.children) {
        loopCNodeListWithCallback(node.children, {
            context: instance.context,
            props: params
        }, childNode => {
            if (typeof childNode === 'string') {
                styles.push(createStyle(selector, { raw: childNode }, instance, params));
            }
            else {
                traverseCNode(childNode, selectorPaths, styles, instance, params);
            }
        });
    }
    selectorPaths.pop();
    // eslint-disable-next-line @typescript-eslint/strict-boolean-expressions
    if ($ && $.after)
        $.after(instance.context);
}
function render(node, instance, props) {
    const styles = [];
    traverseCNode(node, [], styles, instance, props);
    return styles.join('\n\n');
}

function removeElement(el) {
    /* istanbul ignore if */
    // eslint-disable-next-line @typescript-eslint/strict-boolean-expressions
    if (!el)
        return;
    const parentElement = el.parentElement;
    /* istanbul ignore else */
    // eslint-disable-next-line @typescript-eslint/strict-boolean-expressions
    if (parentElement)
        parentElement.removeChild(el);
}
function queryElement(id) {
    return document.querySelector(`style[cssr-id="${id}"]`);
}
function createElement(id) {
    const el = document.createElement('style');
    el.setAttribute('cssr-id', id);
    return el;
}

function getCount(el) {
    const count = el.getAttribute('mount-count');
    if (count === null)
        return null;
    return Number(count);
}
function setCount(el, count) {
    el.setAttribute('mount-count', String(count));
}
function unmount(intance, node, id, count) {
    const { els } = node;
    // If target is undefined, unmount all styles
    if (id === undefined) {
        els.forEach(removeElement);
        node.els = [];
    }
    else {
        const target = queryElement(id);
        // eslint-disable-next-line @typescript-eslint/strict-boolean-expressions
        if (target && els.includes(target)) {
            const mountCount = getCount(target);
            if (!count) {
                if (mountCount !== null) {
                    console.error(`[css-render/unmount]: The style with target='${id}' is mounted in no-count mode.`);
                }
                else {
                    removeElement(target);
                    node.els = els.filter(el => el !== target);
                }
            }
            else {
                if (mountCount === null) {
                    console.error(`[css-render/unmount]: The style with target='${id}' is mounted in count mode.`);
                }
                else {
                    if (mountCount <= 1) {
                        removeElement(target);
                        node.els = els.filter(el => el !== target);
                    }
                    else
                        setCount(target, mountCount - 1);
                }
            }
        }
    }
}
function addElementToList(els, target) {
    els.push(target);
}
function mount(instance, node, id, props, count) {
    let target = null;
    const { els } = node;
    let style;
    if (id === undefined) {
        style = node.render(props);
        id = hash__default['default'](style);
    }
    target = queryElement(id);
    if (target === null) {
        target = createElement(id);
        if (style === undefined)
            style = node.render(props);
        target.textContent = style;
        document.head.appendChild(target);
        if (count) {
            setCount(target, 1);
        }
        addElementToList(els, target);
    }
    else {
        const mountCount = getCount(target);
        if (count) {
            if (mountCount === null) {
                console.error(`[css-render/mount]: The style with id='${id}' has been mounted in no-count mode.`);
            }
            else {
                setCount(target, mountCount + 1);
            }
        }
        else {
            if (mountCount !== null) {
                console.error(`[css-render/mount]: The style with id='${id}' has been mounted in count mode.`);
            }
        }
    }
    return target;
}

function wrappedRender(props) {
    return render(this, this.instance, props);
}
const inNode = typeof window === 'undefined';
function wrappedMount(options = {}) {
    /* istanbul ignore next */
    // eslint-disable-next-line @typescript-eslint/strict-boolean-expressions
    if (inNode)
        return {};
    const { target, id } = options;
    const { props, count = false } = options;
    const targetElement = mount(this.instance, this, id !== null && id !== void 0 ? id : target, props, count);
    return targetElement;
}
function wrappedUnmount(options = {}) {
    /* istanbul ignore next */
    // eslint-disable-next-line @typescript-eslint/strict-boolean-expressions
    if (inNode)
        return;
    const { id, target, delay = 0, count = false } = options;
    if (delay === 0)
        unmount(this.instance, this, id !== null && id !== void 0 ? id : target, count);
    else {
        setTimeout(() => unmount(this.instance, this, id !== null && id !== void 0 ? id : target, count), delay);
    }
}
const createCNode = function (instance, $, props, children) {
    return {
        instance,
        $,
        props,
        children,
        els: [],
        render: wrappedRender,
        mount: wrappedMount,
        unmount: wrappedUnmount
    };
};
const c = function (instance, $, props, children) {
    if (Array.isArray($)) {
        return createCNode(instance, { $: null }, null, $);
    }
    if (Array.isArray(props)) {
        return createCNode(instance, $, null, props);
    }
    else if (Array.isArray(children)) {
        return createCNode(instance, $, props, children);
    }
    else {
        return createCNode(instance, $, props, null);
    }
};

function CssRender(config = {}) {
    const cssr = {
        c: ((...args) => c(cssr, ...args)),
        use: (plugin, ...args) => plugin.install(cssr, ...args),
        find: queryElement,
        context: {},
        config
    };
    return cssr;
}

/* istanbul ignore file */

Object.defineProperty(exports, 'hash', {
    enumerable: true,
    get: function () {
        return hash__default['default'];
    }
});
exports.CssRender = CssRender;
exports.default = CssRender;
